import {Endpoint, EndpointOperations} from './endpoint';
import {Operation} from './operation';
import {NamedSet, PredicateMap} from './common';
import {FALLBACK_LANGUAGE, Language} from './language';
import {FALLBACK_MEDIA_TYPE, MediaType} from './media-type';
import {Charset, FALLBACK_CHARSET} from './charset';

export interface BaseAppState {
	endpoints: unknown;
	operations: unknown;
}

export type AppOperations<T extends App> = (
	T extends App<string, infer R>
		? R extends BaseAppState
			? R['operations'] extends []
				? R['operations'] extends readonly string[]
					? R['operations'][number]
					: string
				: string
			: never
		: never
	);

export interface App<AppName extends string = string, AppState extends BaseAppState = BaseAppState> {
	name: AppName;
	endpoints: NamedSet<Endpoint>;
	operations: NamedSet<Operation>;
	languages: NamedSet<Language>;
	mediaTypes: NamedSet<MediaType>;
	charsets: NamedSet<Charset>;
	// todo add stateful types with these methods
	language(language: Language): this;
	mediaType(mediaType: MediaType): this;
	charset(charset: Charset): this;
	operation<NewOperation extends Operation>(newOperation: NewOperation): App<
		AppName,
		{
			endpoints: AppState['endpoints'],
			operations: AppState['operations'] extends Array<unknown> ? [...AppState['operations'], NewOperation['name']] : [NewOperation['name']]
		}
	>;
	endpoint<NewEndpoint extends Endpoint = Endpoint>(
		newEndpoint: EndpointOperations<NewEndpoint> extends AppOperations<this>
			? NewEndpoint
			: never
	): App<
		AppName,
		{
			endpoints: AppState['endpoints'] extends Array<unknown> ? [...AppState['endpoints'], NewEndpoint] : [NewEndpoint],
			operations: AppState['operations']
		}
	>;
}

interface AppParams<Name extends string = string> {
	name: Name;
}

class AppInstance<Params extends AppParams, State extends BaseAppState> implements App<Params['name'], State> {
	readonly name: Params['name'];
	readonly endpoints: NamedSet<Endpoint>;
	readonly operations: NamedSet<Operation>;
	readonly languages: NamedSet<Language>;
	readonly mediaTypes: NamedSet<MediaType>;
	readonly charsets: NamedSet<Charset>;

	constructor(params: Params) {
		this.name = params.name;
		this.endpoints = new Map<Endpoint['name'], Endpoint>();
		this.operations = new PredicateMap<Operation['name'], Operation>((newOperation, s) => (
			s.method === newOperation.method
		));
		this.languages = new Map<Language['name'], Language>([
			[FALLBACK_LANGUAGE.name.toLowerCase(), FALLBACK_LANGUAGE],
		]);
		this.mediaTypes = new Map<MediaType['name'], MediaType>([
			[FALLBACK_MEDIA_TYPE.name.toLowerCase(), FALLBACK_MEDIA_TYPE],
		]);
		this.charsets = new Map<Charset['name'], Charset>([
			[FALLBACK_CHARSET.name.toLowerCase(), FALLBACK_CHARSET],
		]);
	}

	language(language: Language): this {
		this.languages.set(language.name.toLowerCase(), language);
		return this;
	}

	mediaType(mediaType: MediaType): this {
		this.mediaTypes.set(mediaType.name.toLowerCase(), mediaType);
		return this;
	}

	charset(charset: Charset): this {
		this.charsets.set(charset.name.toLowerCase(), charset);
		return this;
	}

	operation<NewOperation extends Operation>(newOperation: NewOperation) {
		this.operations.set(newOperation.name.toLowerCase(), newOperation);
		return this;
	}

	endpoint<NewEndpoint extends Endpoint = Endpoint>(newEndpoint: NewEndpoint) {
		const nameNormalized = newEndpoint.name.toLowerCase()
		if (this.endpoints.has(nameNormalized)) {
			throw new Error(`Cannot add duplicate endpoint with name: ${nameNormalized}`);
		}

		this.endpoints.set(nameNormalized, newEndpoint);
		return this;
	}
}

export const app = <Params extends AppParams>(params: Params): App<Params['name'], {
	endpoints: [];
	operations: [];
}> => {
	return new AppInstance(params);
};
