import {
	App as BaseApp,
	AVAILABLE_EXTENSION_METHODS,
	Endpoint, EndpointQueue,
	GetEndpointParams,
	Operation, serializeEndpointQueue,
	ServiceParams,
	FALLBACK_LANGUAGE,
	FALLBACK_MEDIA_TYPE,
	FALLBACK_CHARSET,
	Language,
	MediaType,
	Charset,
} from '@modal-sh/yasumi';
import {Client, ClientParams, ClientConnection} from '@modal-sh/yasumi/client';

declare module '@modal-sh/yasumi/client' {
	interface ClientConnection {
		host: string;
		port: number;
		basePath: string;
	}
}

const DEFAULT_HOST = '0.0.0.0' as const;

const DEFAULT_PORT = 80 as const;

const DEFAULT_METHOD = 'GET' as const;

const EXTENSION_METHOD_EFFECTIVE_METHOD = 'POST' as const;

class ClientInstance<App extends BaseApp> implements Client<App> {
	readonly app: App;
	private readonly fetchFn: typeof fetch;
	private connection?: ServiceParams;
	private endpointQueue = [] as EndpointQueue;
	private requestLanguage?: Language;
	private requestMediaType?: MediaType;
	private requestCharset?: Charset;
	private responseLanguage?: Language;
	private responseMediaType?: MediaType;
	private responseCharset?: Charset;

	constructor(params: ClientParams<App>) {
		this.app = params.app;
		this.fetchFn = params.fetch ?? fetch;
		this.requestLanguage = params.app.languages.get(FALLBACK_LANGUAGE.name.toLowerCase());
		this.requestMediaType = params.app.mediaTypes.get(FALLBACK_MEDIA_TYPE.name.toLowerCase());
		this.requestCharset = params.app.charsets.get(FALLBACK_CHARSET.name.toLowerCase());

		this.responseLanguage = params.app.languages.get(FALLBACK_LANGUAGE.name.toLowerCase());
		this.responseMediaType = params.app.mediaTypes.get(FALLBACK_MEDIA_TYPE.name.toLowerCase());
		this.responseCharset = params.app.charsets.get(FALLBACK_CHARSET.name.toLowerCase());
	}

	async connect(params: ServiceParams): Promise<ClientConnection> {
		// we should be explicit with connection params in order not to give a false impression to the user they are "connected"
		// to the target server
		const connection = {
			host: params.host ?? DEFAULT_HOST,
			port: params.port ?? DEFAULT_PORT,
			basePath: params.basePath ?? '',
		};

		this.connection = connection;

		return connection;
	}

	async disconnect() {
		// noop
	}

	at<TheEndpoint extends Endpoint = Endpoint>(endpoint: TheEndpoint, params?: Record<GetEndpointParams<TheEndpoint>, unknown>) {
		if (Array.isArray(this.endpointQueue)) {
			this.endpointQueue?.push([endpoint, params]);
		}
		return this;
	}

	makeRequest(operation: Operation) {
		const baseUrlFragments = [
			this.connection?.host ?? DEFAULT_HOST
		];

		const thePort = (this.connection?.port ?? DEFAULT_PORT);
		if (thePort !== DEFAULT_PORT) {
			baseUrlFragments.push(thePort.toString());
		}

		// TODO how to set to https?
		const scheme = 'http';
		// todo need a way to decode url back to endpoint queue
		const urlString = serializeEndpointQueue(this.endpointQueue);
		this.endpointQueue = [];

		const url = new URL(
			this.connection?.basePath ? `${this.connection.basePath}${urlString}` : urlString,
			`${scheme}://${baseUrlFragments.join(':')}`
		);
		if (typeof operation.searchParams !== 'undefined') {
			url.search = operation.searchParams.toString();
		}
		const rawEffectiveMethod = (operation.method ?? DEFAULT_METHOD).toUpperCase();
		const finalEffectiveMethod = (AVAILABLE_EXTENSION_METHODS as unknown as string[]).includes(rawEffectiveMethod)
			? EXTENSION_METHOD_EFFECTIVE_METHOD
			: rawEffectiveMethod;

		const effectiveResponseMediaType = this.responseMediaType ?? FALLBACK_MEDIA_TYPE;
		const effectiveResponseCharset = this.responseCharset ?? FALLBACK_CHARSET;
		const effectiveResponseLanguage = this.responseLanguage ?? FALLBACK_LANGUAGE;

		if (typeof operation.body !== 'undefined') {
			const effectiveRequestMediaType = this.requestMediaType ?? FALLBACK_MEDIA_TYPE;
			const bodySerialized = effectiveRequestMediaType.serialize(operation.body);

			const effectiveRequestCharset = this.requestCharset ?? FALLBACK_CHARSET;
			const bodyEncoded = effectiveRequestCharset.encode(bodySerialized);

			const effectiveRequestLanguage = this.requestLanguage ?? FALLBACK_LANGUAGE;

			return this.fetchFn(
				url,
				{
					method: finalEffectiveMethod,
					body: bodyEncoded,
					headers: {
						// TODO add wildcard for accept headers to accept anything from the server
						'Accept': effectiveResponseMediaType.name,
						'Accept-Charset': effectiveResponseCharset.name,
						'Accept-Language': effectiveResponseLanguage.name,
						'Content-Language': effectiveRequestLanguage.name,
						'Content-Type': `${effectiveRequestMediaType.name};charset=${effectiveRequestCharset.name}`,
					},
				},
			);
		}

		return this.fetchFn(
			url,
			{
				method: finalEffectiveMethod,
				headers: {
					'Accept': effectiveResponseMediaType.name,
					'Accept-Charset': effectiveResponseCharset.name,
					'Accept-Language': effectiveResponseLanguage.name,
				},
			},
		);
	}
}

export const client = <App extends BaseApp>(params: ClientParams<App>): Client<App> => {
	return new ClientInstance(params);
};
