import {DataSource, ImplementationFunction} from '@modal-sh/yasumi/backend';
import {operation as defineOperation, validation as v} from '@modal-sh/yasumi';
import {
	DataSourceMethodNotImplementedResponseError, IncompleteEndpointMetadataResponseError,
	ResourceCreatedResponse,
	UnableToAssignIdFromDataSourceResponseError, UnableToCreateResourceResponseError,
} from '../response';

export const name = 'create' as const;

export const method = 'POST' as const;

export const operation = defineOperation({
	name,
	method,
});

export const implementation: ImplementationFunction = async (ctx) => {
	const { setLocation, language, endpoint, dataSource, body } = ctx;
	const effectiveDataSource: DataSource = endpoint.dataSource ?? dataSource ?? {} as DataSource;
	const { newId, create, getTotalCount } = effectiveDataSource;

	const idAttr = endpoint.metadata.get('idAttr'); // todo use metadata
	if (typeof idAttr !== 'string') {
		throw new IncompleteEndpointMetadataResponseError(
			// TODO get the status message
		);
	}

	if (typeof newId === 'undefined') {
		throw new DataSourceMethodNotImplementedResponseError(language.statusMessages.dataSourceMethodNotImplemented);
	}

	if (typeof create === 'undefined') {
		throw new DataSourceMethodNotImplementedResponseError(language.statusMessages.dataSourceMethodNotImplemented);
	}

	let resourceNewId;
	let params: v.Output<typeof endpoint.schema>;
	try {
		resourceNewId = await newId();
		params = { ...body as Record<string, unknown> };
		params[idAttr] = resourceNewId;
	} catch (cause) {
		throw new UnableToAssignIdFromDataSourceResponseError(
			language.statusMessages.unableToAssignIdFromResourceDataSource,
			{
				cause,
			}
		);
	}

	setLocation(`/${endpoint.name}/${resourceNewId}`);

	let newObject;
	let totalItemCount: number | undefined;
	// TODO put this in ImplementationContext argument
	const showTotalItemCountOnCreateItem = false;
	try {
		if (
			showTotalItemCountOnCreateItem
			&& typeof getTotalCount === 'function'
		) {
			totalItemCount = await getTotalCount();
			totalItemCount += 1;
		}
		newObject = await create(params);
	} catch (cause) {
		throw new UnableToCreateResourceResponseError(
			language.statusMessages.unableToCreateResource,
			{
				cause,
			}
		);
	}

	const headers: Record<string, string> = {};

	// no need to set location here
	if (typeof totalItemCount !== 'undefined') {
		headers['X-Resource-Total-Item-Count'] = totalItemCount.toString();
	}

	return new ResourceCreatedResponse({
		statusMessage: language.statusMessages.resourceCreated,
		body: newObject,
		headers,
	});
};
